==> 0.1.0.tar.gz: lib/src/dispatcher_builder.dart <==
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/dart/element/visitor.dart';
import 'package:build/build.dart';
import 'package:controller/controller.dart';
import 'package:source_gen/source_gen.dart';

const httpRequestMeta = {
  'Get',
  'Put',
  'Post',
  'Delete',
  'Options',
  'Patch',
  'Head',
  'HttpRequest',
};

class DispatcherBuilder extends GeneratorForAnnotation<Controller> {
  @override
  String generateForAnnotatedElement(Element element, ConstantReader annotation, BuildStep buildStep) {
    if (element.kind != ElementKind.CLASS) {
      throw 'ERROR: @Constructor can only be used on a class, found on $element';
    }
    var typeName = element.name;
    var matcherBuilder = MatcherBuilder();
    element.visitChildren(matcherBuilder);

    if (matcherBuilder.matchers.isEmpty) {
      return null;
    }
    return '''
    class ${typeName}\$DispatcherBuilder implements DispatcherBuilder {
      final ${typeName} controller;

      ProductionController\$DispatcherBuilder(this.controller);
      
      @override
      AsyncPatternMatcher<Request, FutureOr<Response>> appendMatchers(
          AsyncPatternMatcher<Request, FutureOr<Response>> requestMatcher) {
        return requestMatcher
        ${matcherBuilder.matchers.join('\n')};
      }
    }
    ''';
  }
}

class MatcherBuilder extends SimpleElementVisitor<void> {
  final List<String> matchers = [];

  @override
  void visitMethodElement(MethodElement element) {
    matchers.add(element.metadata
        .map(_extractHttpRequest)
        .where((element) => element != null)
        .map((httpRequest) => _buildRequestHandler(element, httpRequest))
        .join('\n'));
  }

  HttpRequest _extractHttpRequest(ElementAnnotation element) {
    var value = element.computeConstantValue();
    if (value.type.element.library.name == 'controller.meta' && httpRequestMeta.contains(value.type.element.name)) {
      return _constructHttpRequest(value);
    }
    return null;
  }

  HttpRequest _constructHttpRequest(DartObject value) {
    if (value.type.element.name == 'HttpRequest') {
      return HttpRequest(value.getField('method').toStringValue(), value.getField('path').toStringValue());
    } else {
      return _constructHttpRequest(value.getField('(super)'));
    }
  }

  String _buildRequestHandler(MethodElement element, HttpRequest httpRequest) {
    var bindArguments = _bindArguments(element);
    var decodeBody = '';
    if (bindArguments.contains("decodedBody['")) {
      decodeBody = 'var decodedBody = jsonDecode(body);';
    }
    return '''
    .when2(
      matchRequest('${httpRequest.method}', '${httpRequest.path}'),
      (Map<String, String> headers, body) {
        $decodeBody
        return controller.${element.name}(${bindArguments});
      },
    )
    ''';
  }

  String _bindArguments(MethodElement element) => '${element.parameters.map(_bindArgument).join(', ')},';

  String _bindArgument(ParameterElement element) => _convertIfNecessary(
        element.type,
        _isBodyProperty(element) ? "decodedBody['${element.name}']" : "headers['${element.name}']",
        element.defaultValueCode,
      );

  String _convertIfNecessary(DartType type, String argument, String defaultValue) {
    if (type.isDartCoreString) {
      return _applyDefault(argument, argument, defaultValue);
    } else if (type.isDartCoreBool) {
      return _applyDefault(argument, 'bool.parse($argument)', defaultValue);
    } else if (type.isDartCoreInt) {
      return _applyDefault(argument, 'int.parse($argument)', defaultValue);
    } else if (type.element.library.name == 'dart.core' && type.element.name == 'DateTime') {
      return _applyDefault(argument, 'DateTime.parse($argument)', defaultValue);
    } else {
      throw 'Unable to convert $type to String';
    }
  }

  String _applyDefault(String argument, String expression, String defaultValue) =>
      argument == expression && defaultValue == null ? argument : '$argument != null ? $expression : $defaultValue';

  bool _isBodyProperty(ParameterElement element) => element.metadata
      .any((element) => _isType(element.computeConstantValue().type, 'controller.meta', 'BodyProperty'));

  bool _isType(DartType type, String library, String name) =>
      type.element.library.name == library && type.element.name == name;
}

==> 0.1.1.tar.gz: lib/src/dispatcher_builder.dart <==
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/dart/element/visitor.dart';
import 'package:build/build.dart';
import 'package:controller/controller.dart';
import 'package:source_gen/source_gen.dart';

const httpRequestMeta = {
  'Get',
  'Put',
  'Post',
  'Delete',
  'Options',
  'Patch',
  'Head',
  'HttpRequest',
};

class DispatcherBuilder extends GeneratorForAnnotation<Controller> {
  @override
  String generateForAnnotatedElement(Element element, ConstantReader annotation, BuildStep buildStep) {
    if (element.kind != ElementKind.CLASS) {
      throw 'ERROR: @Constructor can only be used on a class, found on $element';
    }
    var typeName = element.name;
    var matcherBuilder = MatcherBuilder();
    element.visitChildren(matcherBuilder);

    if (matcherBuilder.matchers.isEmpty) {
      return null;
    }
    return '''
    class ${typeName}\$DispatcherBuilder implements DispatcherBuilder {
      final ${typeName} controller;
      ${matcherBuilder.hasSecurity ? 'final Security security;' : ''}

      ${typeName}\$DispatcherBuilder(this.controller${matcherBuilder.hasSecurity ? ', this.security' : ''});
      
      @override
      AsyncPatternMatcher<Request, FutureOr<Response>> appendMatchers(
          AsyncPatternMatcher<Request, FutureOr<Response>> requestMatcher) {
        return requestMatcher
        ${matcherBuilder.matchers.join('\n')};
      }
    }
    ''';
  }
}

class MatcherBuilder extends SimpleElementVisitor<void> {
  final List<String> matchers = [];
  bool hasSecurity = false;

  @override
  void visitMethodElement(MethodElement element) {
    var httpRequests = element.metadata.map(_extractHttpRequest).where((httpRequest) => httpRequest != null);
    var security = _getMetadataOfType(element.metadata, 'Secured',
        (value) => Secured(value.getField('claims').toListValue().map((claim) => claim.toStringValue()).toList()));
    if (security != null) {
      hasSecurity = true;
    }

    matchers.add(httpRequests.map((httpRequest) => _buildRequestHandler(element, httpRequest, security)).join('\n'));
  }

  T _getMetadataOfType<T>(List<ElementAnnotation> metadata, String name, T Function(DartObject object) constructor) =>
      metadata.map((meta) {
        var value = meta.computeConstantValue();
        if (value.type.element.library.name == 'controller.meta' && value.type.element.name == name) {
          return constructor(value);
        }
        return null;
      }).firstWhere((element) => element != null, orElse: () => null);

  HttpRequest _extractHttpRequest(ElementAnnotation element) {
    var value = element.computeConstantValue();
    if (value.type.element.library.name == 'controller.meta' && httpRequestMeta.contains(value.type.element.name)) {
      return _constructHttpRequest(value);
    }
    return null;
  }

  HttpRequest _constructHttpRequest(DartObject value) {
    if (value.type.element.name == 'HttpRequest') {
      return HttpRequest(value.getField('method').toStringValue(), value.getField('path').toStringValue());
    } else {
      return _constructHttpRequest(value.getField('(super)'));
    }
  }

  String _buildRequestHandler(MethodElement element, HttpRequest httpRequest, Secured security) {
    var bindArguments = _bindArguments(element);
    var decodeBody = '';
    if (bindArguments.contains("decodedBody['")) {
      decodeBody = 'var decodedBody = jsonDecode(body);';
    }
    var checkSecurity = security != null
        ? '''
        if (!security.verify(headers, [${security.claims.map((claim) => "'$claim'").join(',')}])) {
          return Response(401);
        }
        '''
        : '';
    return '''
    .when2(
      matchRequest('${httpRequest.method}', '${httpRequest.path}'),
      (Map<String, String> headers, body) {
        $checkSecurity
        $decodeBody
        return controller.${element.name}(${bindArguments});
      },
    )
    ''';
  }

  String _bindArguments(MethodElement element) => '${element.parameters.map(_bindArgument).join(', ')}';

  String _bindArgument(ParameterElement element) => _convertIfNecessary(
        element.type,
        _isBodyProperty(element) ? "decodedBody['${element.name}']" : "headers['${element.name}']",
        element.defaultValueCode,
      );

  String _convertIfNecessary(DartType type, String argument, String defaultValue) {
    if (type.isDartCoreString) {
      return _applyDefault(argument, argument, defaultValue);
    } else if (type.isDartCoreBool) {
      return _applyDefault(argument, 'bool.parse($argument)', defaultValue);
    } else if (type.isDartCoreInt) {
      return _applyDefault(argument, 'int.parse($argument)', defaultValue);
    } else if (type.element.library.name == 'dart.core' && type.element.name == 'DateTime') {
      return _applyDefault(argument, 'DateTime.parse($argument)', defaultValue);
    } else {
      throw 'Unable to convert $type to String';
    }
  }

  String _applyDefault(String argument, String expression, String defaultValue) =>
      argument == expression && defaultValue == null ? argument : '$argument != null ? $expression : $defaultValue';

  bool _isBodyProperty(ParameterElement element) => element.metadata
      .any((element) => _isType(element.computeConstantValue().type, 'controller.meta', 'BodyProperty'));

  bool _isType(DartType type, String library, String name) =>
      type.element.library.name == library && type.element.name == name;
}
